from dataclasses import dataclass
from typing import Any, List, Type, TypeVar, Union

from aiogram import Bot

from axaiogram.static import MessagesGroup

from .template_group import TemplateGroup
from .template_parser import TemplateParser


@dataclass
class MessagesCollection:
    _groups = {}
    _lang_code: str = "en"

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)

        except AttributeError:
            if __name not in self._groups:

                cls = self.__class__.__annotations__.get(__name)

                if cls and issubclass(MessagesGroup, cls):
                    self._groups[__name] = cls.from_file(self._lang_code)
                else:
                    raise

            return self._groups[__name]


@dataclass
class TemplatesCollection:
    langs: tuple[str] = ("en",)
    _groups = None

    def __getattribute__(self, __name: str) -> Any:
        try:
            return super().__getattribute__(__name)

        except AttributeError:
            if not self._groups:
                self._groups = {}

            if __name not in self._groups:
                cls = self.__class__.__annotations__.get(__name)

                if cls is not None and issubclass(TemplateGroup, cls):
                    self._groups[__name] = cls.load(self.langs)

                else:
                    raise

            return self._groups[__name]


TTemplateCollection = TypeVar("TTemplateCollection", bound=TemplatesCollection)
TMessagesCollection = TypeVar("TMessagesCollection", bound=MessagesCollection)


class TemplatesHolder:
    def __init__(
        self,
        bot: Bot,
        tc: TemplatesCollection,
        mc: MessagesCollection,
        lang_code: str = "en",
        chat_id: Union[int, None] = None,
    ) -> None:
        self.bot = bot
        self.lang_code = lang_code

        self.collection = tc
        self.msg_collection = mc

        self.chat_id = chat_id

    async def render(
        self,
        template_parser: TemplateParser,
        context: Union[dict, None] = None,
        **kwargs,
    ):
        if context is None:
            context = {}

        res = template_parser.render(
            self.lang_code,
            context={
                "mc": self.msg_collection,
            },
        )

        method = (
            self.bot.send_message
            if res.type_ == "text"
            else self.bot.__getattribute__(f"send_{res.type_}")
        )

        if not kwargs.get("chat_id"):
            kwargs["chat_id"] = self.chat_id

        return await method(**res.to_message_data(**kwargs))


class TemplateFactory:
    def __init__(
        self,
        bot: Bot,
        tc: Type[TTemplateCollection],
        mc: Type[TMessagesCollection],
        langs: Union[List[str], None] = None,
    ) -> None:

        if not langs:
            raise NotImplementedError("langs aren't set up")

        self.bot = bot
        self.langs: List[str] = langs

        self.msg_collection_type = mc
        self.temp_collection = tc(langs=self.langs)

        self._msg_collections = {}

    def get_template(
        self, lang_code: str, chat_id: Union[int, None] = None
    ) -> TemplatesHolder:

        if lang_code not in self.langs:
            raise NotImplementedError("Lang code not found")

        if lang_code not in self._msg_collections:
            self._msg_collections[lang_code] = self.msg_collection_type(
                _lang_code=lang_code
            )

        return TemplatesHolder(
            bot=self.bot,
            tc=self.temp_collection,
            mc=self._msg_collections[lang_code],
            lang_code=lang_code,
            chat_id=chat_id,
        )



from .auth.on_the_fly_user_auth_middleware import OnTheFlyUserAuthMiddleware
from .template_middleware import TemplateMiddleWare

__all__ = [
    "TemplateMiddleWare",
    "OnTheFlyUserAuthMiddleware",
]



import random
from uuid import uuid4

from aiogram.utils.keyboard import InlineKeyboardBuilder, ReplyKeyboardBuilder

from templates.renderers.md.template_parser import MarkdownTemplateParser


class TestMarkdownTemplateParser:
    def test_simple_text_one_lang(self):
        base_format = """
        some cool test
        """

        parser = MarkdownTemplateParser({"en": base_format})

        res = parser.render("en")

        assert res is not None
        assert res.text is not None
        assert res.text.strip() == base_format.strip()

    def test_simple_text_two_lang(self):
        base_format_en = """
        some cool test
        """
        base_format_ru = """
        сроки жмут слышком
        """

        parser = MarkdownTemplateParser(
            {
                "en": base_format_en,
                "ru": base_format_ru,
            }
        )

        res = parser.render("en")

        assert res is not None
        assert res.text is not None

        assert res.text.strip() == base_format_en.strip()

        res = parser.render("ru")

        assert res is not None
        assert res.text is not None

        assert res.text.strip() == base_format_ru.strip()

    def test_simple_one_reply_keyboard(self):
        key_text = "only one reply key"
        base_format = f"""
        something cool to say1

        | !reply |
        | [{key_text}]() |
        """

        parser = MarkdownTemplateParser({"en": base_format})

        res = parser.render("en")

        assert type(res.keyboard) is ReplyKeyboardBuilder
        keys = list(res.keyboard.buttons)

        assert len(keys) == 1

        assert list(keys)[0].text == key_text

        assert res.keyboard is not None

    def test_row_reply_keyboard(self):
        key_text1 = str(random.randint(1, 9999999))
        key_text2 = str(random.randint(1, 9999999))

        base_format = f"""
        something cool to say1

        | !reply |
        | [{key_text1}]() | [{key_text2}]() |
        """

        parser = MarkdownTemplateParser({"en": base_format})

        res = parser.render("en")

        assert type(res.keyboard) is ReplyKeyboardBuilder
        keys = list(res.keyboard.buttons)

        assert len(keys) == 2

        assert list(keys)[0].text == key_text1
        assert list(keys)[1].text == key_text2

        assert res.keyboard is not None

    def test_mutli_row_column_keyboard(self):
        key_text1 = str(random.randint(1, 9999999))
        key_text2 = str(random.randint(1, 9999999))
        key_text3 = str(random.randint(1, 99999999))
        key_text4 = str(random.randint(1, 99999999))
        key_text5 = str(random.randint(1, 99999999))

        base_format = f"""
        something cool to say1

        | !reply |
        | [{key_text1}]() | [{key_text2}]() |
        | [{key_text3}]() | [{key_text4}]() |
        | [{key_text5}]() |
        """

        parser = MarkdownTemplateParser({"en": base_format})

        res = parser.render("en")

        assert type(res.keyboard) is ReplyKeyboardBuilder
        keys = list(res.keyboard.buttons)

        assert len(keys) == 5

        assert list(keys)[0].text == key_text1
        assert list(keys)[1].text == key_text2
        assert list(keys)[2].text == key_text3
        assert list(keys)[3].text == key_text4
        assert list(keys)[4].text == key_text5

        assert res.keyboard is not None

    def test_meta_type(self):
        base_format = """
        something cool to say1

        !meta:
            type: audio

        """

        parser = MarkdownTemplateParser({"en": base_format})

        res = parser.render("en")

        assert res.type_ == "audio"

    def test_data_content(self):
        file_id = str(uuid4())

        base_format = f"""
        something cool to say1

        !meta:
            data-content:
                file_id: {file_id}

        """

        parser = MarkdownTemplateParser({"en": base_format})

        res = parser.render("en")

        assert res.content_data is not None
        assert res.content_data.get("file_id") == file_id

    def test_full_template(self):
        file_id = str(uuid4())
        buttons_text = ""
        buttons = []

        for _ in range(random.randint(1, 8)):
            button_text_row = ""

            for __ in range(random.randint(1, 5)):
                text = str(uuid4())
                data = f"!{str(uuid4())}"

                buttons.append((text, data))
                button_text_row += f" [{text}]({data}) |"

            buttons_text += f"\n{button_text_row}"

        text = str(random.randint(1, 99999999999999))
        base_format = (
            text
            + "\n"
            + "\n | !inline |"
            + buttons_text
            + "\n"
            + f"""
            !meta:
                type: video
                data-content:
                    file_id: {file_id}

            """
        )

        parser = MarkdownTemplateParser(base_formats={"en": base_format})

        res = parser.render("en")

        assert res is not None

        assert res.text is not None
        assert res.text.strip() == text

        assert res.type_ == "video"

        assert type(res.keyboard) is InlineKeyboardBuilder
        keys = list(res.keyboard.buttons)

        assert len(keys) == len(buttons)

        for button, key in zip(buttons, keys):
            assert key.text == button[0]
            assert key.callback_data == button[1]



from abc import ABC, abstractmethod
from typing import Any, Awaitable, Callable, Dict

from aiogram import types
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.fsm.context import FSMContext
from aiogram.fsm.state import StatesGroup
from aiogram.types import TelegramObject
from axsqlalchemy.uow import UOWFactory
from db.repository.collection import RepoCollector

from axaiogram.handlers.frozen_handler import FrozenHandler


class BaseUserAuthHandlers(ABC):
    async def require(self, chat_id: int, state: FSMContext, data: dict):
        raise NotImplementedError


class UseHandlerUserAuthMiddleware(BaseMiddleware):
    def __init__(
        self,
        uowf: UOWFactory[RepoCollector],
        auth_handlers_group: BaseUserAuthHandlers,
        auth_states_group: StatesGroup,
    ) -> None:
        self.uowf = uowf
        self.auth_handlers = auth_handlers_group
        self.auth_states_group = auth_states_group

    async def __call__(
        self,
        handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
        event: types.Message,
        data: Dict[str, Any],
    ) -> Any:
        if not hasattr(event, "from_user"):
            raise NotImplementedError(
                f"`{self.__class__.__name__}` can't handler event type of `{type(event)}`"
            )

        assert event.from_user is not None
        user_id = event.from_user.id
        data["user_id"] = user_id

        async with self.uowf() as uow:
            user = await uow.repo.user.get(user_id)

        if not user or not bool(user.is_active):
            state: FSMContext = data["state"]

            current_state = await state.get_state()
            print(
                f"state in auth states group: {current_state in self.auth_states_group.__states__}"
            )

            if current_state not in self.auth_states_group.__states__:
                await FrozenHandler(handler, event, data).freeze(state)

                return await self.auth_handlers.require(user_id, state, data)

        data["user"] = user

        return await handler(event, data)



from typing import Callable, Dict, Union

from aiogram import Bot, Dispatcher, F, types
from aiogram.fsm.context import FSMContext
from aiogram.utils.keyboard import InlineKeyboardBuilder
from axsqlalchemy.uow import TRepoCollector, UOWFactory

from axaiogram.db.repository.collection import RepoCollector
from axaiogram.handlers.frozen_handler import FrozenHandler

from ..base import BaseHandlersGroup
from ..base import proxy_router as _router


class UserLanguageSetupHandlers(BaseHandlersGroup):
    def __init__(
        self,
        bot: Bot,
        dp: Dispatcher,
        uowf: UOWFactory[TRepoCollector],
        languages: Union[Dict[str, str], None] = None,
        start_handler: Union[Callable, None] = None,
    ) -> None:
        super().__init__(bot, dp, uowf)
        self.uowf = uowf

        if not languages:
            languages = {
                "en": "English",
            }

        self.languages = languages
        self.start_handler = start_handler

    async def _require(self, chat_id: int):
        builder = InlineKeyboardBuilder()

        for code, name in self.languages.items():
            builder.button(text=name, callback_data=f"chlang:{code}")

        builder.adjust(1)

        msg = await self.bot.send_message(
            chat_id=chat_id,
            text="...",
            reply_markup=types.ReplyKeyboardRemove(remove_keyboard=True),
        )

        await msg.delete()

        await self.bot.send_message(
            chat_id=chat_id,
            text="Выберите язык/Choose a language.",
            reply_markup=builder.as_markup(),
        )

    @_router.message(F.text == "/change_lang")
    async def show_list(self, message: types.Message):
        return await self._require(message.chat.id)

    @_router.message(F.text.strip().startswith("🌍"))
    async def handle_list(self, message: types.Message):
        assert message.from_user is not None
        return await self._require(message.from_user.id)

    @_router.callback_query(F.data.startswith("chlang:"))
    async def handle_choosen(self, cq: types.CallbackQuery, state: FSMContext):
        assert cq.data is not None
        assert cq.from_user is not None

        if cq.message:
            await cq.message.delete()

        _, lang_code = cq.data.split(":")

        async with self.uowf() as uow:
            await uow.repo.user_lang.add(
                uow.repo.user_lang.IModel(
                    id=cq.from_user.id,
                    lang_code=lang_code,
                )
            )

        if fhandler := await FrozenHandler.get(state):
            await fhandler.delete(state)

            fhandler.data["lang_code"] = lang_code

            return await fhandler.handler(fhandler.event, fhandler.data)

        if self.start_handler is not None:
            await self.start_handler(cq.from_user.id)



from abc import ABC, abstractmethod
from typing import Dict, Union

from .template_message import TemplateMessage


class TemplateParser(ABC):
    def __init__(self, base_formats: Dict[str, str]) -> None:
        """
        base_formats: Dict[lang_code: "not parsed text"]
        """
        self.base_formats = base_formats

    def _format_text(self, text: str, context: dict) -> str:
        return text.format(**context)

    @abstractmethod
    def render(
        self,
        lang_code: str,
        context: Union[dict, None] = None,
        is_prerender: bool = False,
    ) -> TemplateMessage:
        raise NotImplementedError



from typing import Any, Awaitable, Callable, Dict, Union

from aiogram import types
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import TelegramObject
from axsqlalchemy.uow import TRepoCollector, UOWFactory

from axaiogram.db.repository.collection import RepoCollector
from axaiogram.handlers.frozen_handler import FrozenHandler

from ..handlers.user import UserLanguageSetupHandlers


class UserLanguageMiddleware(BaseMiddleware):
    def __init__(
        self, uowf: UOWFactory[TRepoCollector], handler: UserLanguageSetupHandlers
    ) -> None:
        self.handler = handler
        self.uowf = uowf
        super().__init__()

    async def __call__(
        self,
        handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
        event: Union[TelegramObject, types.Message],
        data: Dict[str, Any],
    ) -> Any:
        user_data: types.User = data["event_from_user"]
        user_id = user_data.id

        async with self.uowf() as uow:
            user_lang_data = await uow.repo.user_lang.get(user_id)

        if not user_lang_data:
            state = data["state"]

            await FrozenHandler(handler, event, data).freeze(state)

            return await self.handler._require(user_id)

        data["lang_code"] = user_lang_data.lang_code

        return await handler(event, data)



from abc import ABC
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, TypeVar, Union

from aiogram import Bot, Dispatcher, Router

from axabc.db import AbstractUOWFactory

TUOW = TypeVar('TUOW', bound=AbstractUOWFactory)

class BaseHandlersGroup(ABC):
    __router: Union[None, Router] = None

    def __init__(
        self,
        bot: Bot,
        dp: Dispatcher,
        uowf: Union[TUOW, None] = None,
    ) -> None:
        self.router = proxy_router.get_group_observed_router(self)

        self.bot = bot
        self.dp = dp
        self.uowf = uowf

    @classmethod
    @property
    def router_collector(cls) -> Router:
        if not cls.__router:
            cls.__router = Router(name=cls.__name__)
        return cls.__router


OBSERVERS_NAME = [
    "message",
    "edited_message",
    "channel_post",
    "edited_channel_post",
    "inline_query",
    "chosen_inline_result",
    "callback_query",
    "shipping_query",
    "pre_checkout_query",
    "poll",
    "poll_answer",
    "my_chat_member",
    "chat_member",
    "chat_join_request",
    "error",
]

@dataclass
class FrozenHandler:
    observer_name: str

    args: Union[List[Any], tuple[Any]]
    kwargs: Dict[str, Any]
    fn: Any


class ProxyRouter(Router):
    _observers_name = OBSERVERS_NAME

    def __init__(self, *, name: str | None = None) -> None:
        self._cls_registrations: Dict[str, Dict[str, FrozenHandler]] = {}
        super().__init__(name=name)

    def proxy_observer(self, __name: str) -> Callable[[Any], Any]:
        def inner_wrapper(*args: Any, **kwargs: Dict[Any, Any]):
            def wrapper(fn: Callable[[Any], Any]) -> Any:
                clsname, methodname, *extra = fn.__qualname__.split('.')
                if not extra:
                    if clsname not in self._cls_registrations:
                        self._cls_registrations[clsname] = {}

                    self._cls_registrations[clsname][methodname] = FrozenHandler(
                        observer_name=__name,
                        args=args,
                        kwargs=kwargs,
                        fn=fn,
                    )

                return fn

            return wrapper

        return inner_wrapper

    def get_group_observed_router(self, group: BaseHandlersGroup) -> Router:
        cls = group.__class__
        clsname = cls.__name__
        router = Router(name=clsname)

        if clsname not in self._cls_registrations:
            raise NotImplementedError(f"{clsname} hasn't registered observers")

        for methodname, fh in self._cls_registrations[clsname].items():
            getattr(router, fh.observer_name)(*fh.args, **fh.kwargs)(getattr(group, methodname))

        return router


    def __getattribute__(self, __name: str) -> Any:
        if not __name.startswith('_') and __name in self._observers_name:
            return self.proxy_observer(__name)
        return super().__getattribute__(__name)


proxy_router = ProxyRouter()



import sqlalchemy as sa
from axsqlalchemy.model import Base, BaseTableAt


class AbstractUserModel(BaseTableAt):
    __abstract__ = True

    id = sa.Column(sa.BigInteger, primary_key=True, index=True)
    username = sa.Column(sa.String(length=256), nullable=True)
    firstname = sa.Column(sa.String(length=456), nullable=False)
    secondname = sa.Column(sa.String(length=456), nullable=True)
    is_active = sa.Column(sa.Boolean, nullable=False, default=True)


class AbstractUserLanguageModel(BaseTableAt):
    __abstract__ = True

    id = sa.Column(sa.BigInteger, primary_key=True, index=True)
    lang_code = sa.Column(sa.String(length=256), nullable=True)


class AbstractUserWithEmailModel(AbstractUserModel):
    __abstract__ = True

    email = sa.Column(sa.String, nullable=False)


__all__ = [
    "Base",
    "AbstractUserModel",
    "AbstractUserWithEmailModel",
]



import json
from pathlib import Path
from typing import Any, Dict, Union

from pydantic import BaseModel


class MessagesGroup(BaseModel):
    path: str
    lang_code: Union[str, None] = None

    @classmethod
    def from_file(cls, lang_code: Union[str, None] = None):
        path = Path(cls.path)

        messages: Dict[str, Any] = cls.__get_messages(path, lang_code)
        messages["lang_code"] = lang_code

        return cls(**messages)

    @classmethod
    def __get_messages(cls, path, lang_code):
        with open(path, "r") as f:
            messages: Dict[str, Any] = json.load(f)

        _messages = {}

        for key in cls.__fields__.keys():
            if key != "lang_code":
                values: Union[Dict[str, Any], str, None] = messages.get(key)

                if values is None:
                    raise NotImplementedError(f"key `{key}` not fould in file `{path}`")

                if lang_code is not None:
                    cls.__get_from_lang(key, values, lang_code, path, _messages)
                else:
                    _messages[key] = values

        return _messages

    @classmethod
    def __get_from_lang(
        cls,
        key: str,
        value: Dict[str, Any],
        lang_code: str,
        path: Path,
        messages: dict,
    ):
        if type(value) is dict:
            langs = value.get("lang")

            if langs is not None and type(langs) is dict:
                lang = langs.get(lang_code)

                if lang is None:
                    raise NotImplementedError(
                        f"lang_code `{lang_code}` is not found in key `{key}` file `{path}`"
                    )
                messages[key] = lang

                return

        raise NotImplementedError(f"key `{key}` not contains lang param")



from typing import Union

from .collection import (MessagesCollection, MessagesGroup, TemplateFactory,
                         TemplateGroup, TemplatesCollection, TemplatesHolder)

TItem = Union[dict, str]

__all__ = [
    "TemplateGroup",
    "MessagesCollection",
    "TemplateFactory",
    "TemplatesHolder",
    "MessagesGroup",
    "TemplatesCollection",
    "TItem",
]



from typing import Callable, Union

from aiogram import types
from aiogram.fsm.context import FSMContext


class FrozenHandler:
    _state_key = "frozen_handler"

    def __init__(
        self, handler: Callable, event: types.TelegramObject, data: dict
    ) -> None:
        self.handler = handler
        self.event = event
        self.data = data

    async def freeze(self, state: FSMContext) -> None:
        data = await state.get_data()

        await state.set_data(
            {
                self._state_key: self,
                **data,
            }
        )

    async def delete(self, state: FSMContext) -> None:
        data = await state.get_data()

        await state.set_data(
            {
                self._state_key: None,
                **data,
            }
        )

    @classmethod
    async def get(cls, state: FSMContext) -> Union["FrozenHandler", None]:
        data = await state.get_data()

        return data.get(cls._state_key)



from typing import Any, Awaitable, Callable, Dict, Union

from aiogram import types
from aiogram.dispatcher.middlewares.base import BaseMiddleware
from aiogram.types import TelegramObject
from axsqlalchemy.uow import UOWFactory
from db.repository.collection import RepoCollector


class OnTheFlyUserAuthMiddleware(BaseMiddleware):
    def __init__(self, uowf: UOWFactory[RepoCollector]) -> None:
        self.uowf = uowf
        super().__init__()

    async def __call__(
        self,
        handler: Callable[[TelegramObject, Dict[str, Any]], Awaitable[Any]],
        event: Union[TelegramObject, types.Message],
        data: Dict[str, Any],
    ) -> Any:
        if hasattr(event, "from_user"):
            user_data: types.User = getattr(event, "from_user")

            if user_data:
                chat_id = user_data.id
                async with self.uowf() as uow:
                    user = await uow.repo.user.get(chat_id)

                    if not user:
                        user = await uow.repo.user.add(
                            uow.repo.user.IModel(
                                id=user_data.id,
                                username=user_data.username,
                                firstname=user_data.first_name,
                                secondname=user_data.last_name,
                                lang_code=data.get("lang_code"),
                            )
                        )

                data["user"] = user

        return await handler(event, data)



from axsqlalchemy.uow import BaseRepoCollector

from .user import UserReposity
from .user_lang import UserLangReposity


class RepoCollector(BaseRepoCollector):
    user: UserReposity
    user_lang: UserLangReposity
